require 'pp'

module NavigationHelpers
	##
	# Convenience method of translating 'login page' to its expanded URL
	#
	# @param [String] page_name
	# @return [String] URL
	# @raise [RuntimeError] Requested the URL of a page name which is not yet defined
	def path_to(page_name)
		"https://#{SERVER}/" +
		case page_name
		when /LogServer main page/
			'logserver/'
		when /main page/
			'monitor/index.php/tac/index'
		when /LogServer login page/
			'logserver'
		when /View logs/
			'logserver/view/logs'
		when /login page/
			'monitor/index.php/default/show_login'
		when /nagvis page/
			'monitor/index.php/nagvis/index'
		when /nacoma page/
			'monitor/op5/nacoma/'
		when /Host details page/
			'monitor/index.php/status/host/all'
		when /Service details page/
			'monitor/index.php/status/service/all'
		when /Configure page/
			'monitor//index.php/configuration/configure'
		when /PNP page/
			'monitor/index.php/pnp/'
		when /"View logs" page/
			'logserver/view/logs/'
		when /list view/
			'/monitor/index.php/listview'
		else
			raise "Can't find mapping from \"#{page_name}\" to a path.\n" +
				"Now, go and add a mapping in #{__FILE__}"
		end
	end

	##
	# @param path [String] URL in the form of /monitor/index.php/kdjb
	# @return [String] Prepended with domain
	def url_for(path)
		"https://#{SERVER}" + path
	end

	##
	# Look for clickable things on the current page (execute Capybara's visit() first)
	# Should exclude invalid links (javascript:void()) and restrict pagination to a few
	# levels.
	#
	# @return [Array] URLs to visit
	def find_urls
		to_visit = []
		#p "Already visited:"
		#p @visited
		page.all('a').each do |link|
			if !link['href'] or !link['href'].match("^/")
				next
			end
			href = url_for(link['href'])
			if !href.match('javascript')

				stripped = href.sub("page=\d+&?", "")
				if href != stripped
					href = stripped
					# allright, we found something paginated.. let's just not go down this path more than.. twice
					to_visit << stripped + stripped.match('?') ? '&page=2' : '?page=2'
				end

				if !@visited.keys.include?(href) &&
					!to_visit.include?(href)
					#p "Gonna add:"
					#p link['href']
					#p "======"
					to_visit << href
				end
			end
			return to_visit
		end
	end

	##
	# @param [Capybara::Poltergeist::Browser] page
	# @return [Array] Usable strings which explains the occured errors
	def validate_page(page)
		errors = []
		@requested_stuff ||= []

		if page.html.empty?
			errors << "No HTML output on "+page.current_url
		end

		# HTTP Requests
		if not page.driver or not page.driver.network_traffic
			errors << "No network traffic on " + page.current_url + ": " + page.inspect
		else
			page.driver.network_traffic.each do |request|
				request.response_parts.map do |response|
					if @requested_stuff.include?(request.url)
						next
					end
					if not response.status
						errors << "#{request.method} #{request.url}: No response status, #{response.inspect}"
					elsif response.redirect_url and response.body_size
						errors << "#{request.method} #{response.status} #{request.url}: Should not have a response body when redirecting"
					elsif response.status >= 400
						errors << "#{request.method} #{response.status} #{request.url}: #{response.status_text}"
					end
					@requested_stuff << request.url
				end
			end
		end

		# @todo detect stack trace
		# @todo see that a string only occurs on pages whose url matches a set of regex filters, specifically: display_name for host/service

		# HTML validation
		errors |= tidy(page.html)
		#pp page.inspect
		#pp page.current_path
		#pp page.current_url

		return errors
	end

	##
	# Starts the spidering business on given URL, manages the queue
	# of which links to visit next
	#
	# @param [String] url
	def spider(url)
		@visited = {}
		#p @to_visit
		@to_visit = [url]
		while url = @to_visit.pop()
			#if @visited.length > 60
			#shut down this spidering business
			#return
			#end
			begin
				visit(url)
			rescue Capybara::Poltergeist::JavascriptError => e
				#p "JS error on #{url}: #{e}"
				@visited[url] = ["JS error on #{url}: #{e}"]
				next
			end
			p "Visiting "+url
			@visited[url] = validate_page(page)
			@to_visit |= find_urls()

			# stored as 'index.php/tac' => [] if no errors,
			# 'index.php/tac' => ['404: Not found'] if errors
			#p "Ok, validated "+url
			#p "Now @visited looks like:"
			#pp @visited
			#p "--------"
		end
	end

	##
	# Print result of spidering business
	#
	# @raise [SpiderError] One or more errors were found
	def react_to_spider_errors
		failed = false
		@visited.keys.each do |link|
			@visited[link].each do |errors|
				errors.each do |error|
					failed = true
					pp link+": "+errors+"\n"
				end
			end
		end
		if failed
			raise SpiderError("The spidering business found some nastyness")
		else
			p "Not a single spidering error, congrats!"
		end
	end
end

After do |scenario|
	if scenario.title == 'All links should pass our custom validation rules'
		react_to_spider_errors()
	end
end

class SpiderError < Exception
end

World(NavigationHelpers)
